@@ -4,25 +4,38 @@
  */
 
 #ifdef HAVE_BUILTIN___BUILTIN_CONSTANT_P
+#define write_warn(str, x) RB_GNUC_EXTENSION_BLOCK( \
+        NIL_P(str) ? \
+            warn_print(x) : (void)( \
+            (__builtin_constant_p(x)) ? 		\
+                rb_str_concat((str), rb_str_new((x), (long)strlen(x))) : \
+                rb_str_concat((str), rb_str_new2(x)) \
+            ) \
+        )
 #define warn_print(x) RB_GNUC_EXTENSION_BLOCK(	\
     (__builtin_constant_p(x)) ? 		\
 	rb_write_error2((x), (long)strlen(x)) : \
 	rb_write_error(x)			\
 )
 #else
+#define write_warn(str, x) NIL_P(str) ? rb_write_error((x)) : (void)rb_str_concat((str), rb_str_new2(x))
 #define warn_print(x) rb_write_error(x)
 #endif
+
+#define write_warn2(str,x,l) NIL_P(str) ? warn_print2(x,l) : (void)rb_str_concat((str), rb_str_new((x),(l)))
 #define warn_print2(x,l) rb_write_error2((x),(l))
+
+#define write_warn_str(str,x) NIL_P(str) ? rb_write_error_str(x) : (void)rb_str_concat((str), (x))
 #define warn_print_str(x) rb_write_error_str(x)
 
 static VALUE error_pos_str(void);
 
 static void
-error_pos(void)
+error_pos(const VALUE str)
 {
-    VALUE str = error_pos_str();
-    if (!NIL_P(str)) {
-	warn_print_str(str);
+    VALUE pos = error_pos_str();
+    if (!NIL_P(pos)) {
+	write_warn_str(str, pos);
     }
 }
 
@@ -32,7 +45,7 @@ error_pos_str(void)
     int sourceline;
     VALUE sourcefile = rb_source_location(&sourceline);
 
-    if (sourcefile) {
+    if (!NIL_P(sourcefile)) {
 	ID caller_name;
 	if (sourceline == 0) {
 	    return rb_sprintf("%"PRIsVALUE": ", sourcefile);
@@ -67,69 +80,55 @@ set_backtrace(VALUE info, VALUE bt)
 }
 
 static void
-error_print(rb_thread_t *th)
+error_print(rb_execution_context_t *ec)
 {
-    rb_threadptr_error_print(th, th->errinfo);
+    rb_ec_error_print(ec, ec->errinfo);
 }
 
-void
-rb_threadptr_error_print(rb_thread_t *volatile th, volatile VALUE errinfo)
+#define CSI_BEGIN "\033["
+#define CSI_SGR "m"
+
+static const char underline[] = CSI_BEGIN"1;4"CSI_SGR;
+static const char bold[] = CSI_BEGIN"1"CSI_SGR;
+static const char reset[] = CSI_BEGIN""CSI_SGR;
+
+static void
+print_errinfo(const VALUE eclass, const VALUE errat, const VALUE emesg, const VALUE str, int colored)
 {
-    volatile VALUE errat = Qundef;
-    volatile int raised_flag = th->raised_flag;
-    volatile VALUE eclass = Qundef, e = Qundef;
-    const char *volatile einfo;
-    volatile long elen;
+    const char *einfo = "";
+    long elen = 0;
     VALUE mesg;
 
-    if (NIL_P(errinfo))
-	return;
-    rb_thread_raised_clear(th);
+    if (emesg != Qundef) {
+	if (NIL_P(errat) || RARRAY_LEN(errat) == 0 ||
+	    NIL_P(mesg = RARRAY_AREF(errat, 0))) {
+	    error_pos(str);
+	}
+	else {
+	    write_warn_str(str, mesg);
+	    write_warn(str, ": ");
+	}
 
-    TH_PUSH_TAG(th);
-    if (TH_EXEC_TAG() == 0) {
-	errat = rb_get_backtrace(errinfo);
-    }
-    else if (errat == Qundef) {
-	errat = Qnil;
-    }
-    else if (eclass == Qundef || e != Qundef) {
-	goto error;
-    }
-    else {
-	goto no_message;
-    }
-    if (NIL_P(errat) || RARRAY_LEN(errat) == 0 ||
-	NIL_P(mesg = RARRAY_AREF(errat, 0))) {
-	error_pos();
-    }
-    else {
-	warn_print_str(mesg);
-	warn_print(": ");
+        if (colored) write_warn(str, bold);
+
+	if (!NIL_P(emesg)) {
+	    einfo = RSTRING_PTR(emesg);
+            elen = RSTRING_LEN(emesg);
+	}
     }
 
-    eclass = CLASS_OF(errinfo);
-    if (eclass != Qundef &&
-	(e = rb_check_funcall(errinfo, rb_intern("message"), 0, 0)) != Qundef &&
-	(RB_TYPE_P(e, T_STRING) || !NIL_P(e = rb_check_string_type(e)))) {
-	einfo = RSTRING_PTR(e);
-	elen = RSTRING_LEN(e);
-    }
-    else {
-      no_message:
-	einfo = "";
-	elen = 0;
-    }
     if (eclass == rb_eRuntimeError && elen == 0) {
-	warn_print("unhandled exception\n");
+        if (colored) write_warn(str, underline);
+	write_warn(str, "unhandled exception\n");
     }
     else {
 	VALUE epath;
 
 	epath = rb_class_name(eclass);
 	if (elen == 0) {
-	    warn_print_str(epath);
-	    warn_print("\n");
+            if (colored) write_warn(str, underline);
+	    write_warn_str(str, epath);
+	    write_warn(str, "\n");
 	}
 	else {
 	    const char *tail = 0;
@@ -141,50 +140,127 @@ rb_threadptr_error_print(rb_thread_t *volatile th, volatile VALUE errinfo)
 		len = tail - einfo;
 		tail++;		/* skip newline */
 	    }
-	    warn_print_str(tail ? rb_str_subseq(e, 0, len) : e);
+	    write_warn_str(str, tail ? rb_str_subseq(emesg, 0, len) : emesg);
 	    if (epath) {
-		warn_print(" (");
-		warn_print_str(epath);
-		warn_print(")\n");
+		write_warn(str, " (");
+                if (colored) write_warn(str, underline);
+                write_warn_str(str, epath);
+                if (colored) write_warn(str, reset);
+                if (colored) write_warn(str, bold);
+		write_warn(str, ")\n");
 	    }
 	    if (tail) {
-		warn_print_str(rb_str_subseq(e, tail - einfo, elen - len - 1));
+		write_warn_str(str, rb_str_subseq(emesg, tail - einfo, elen - len - 1));
 	    }
-	    if (tail ? einfo[elen-1] != '\n' : !epath) warn_print2("\n", 1);
+	    if (tail ? einfo[elen-1] != '\n' : !epath) write_warn2(str, "\n", 1);
 	}
     }
+    if (colored) write_warn(str, reset);
+}
 
+static void
+print_backtrace(const VALUE eclass, const VALUE errat, const VALUE str, int reverse)
+{
     if (!NIL_P(errat)) {
 	long i;
 	long len = RARRAY_LEN(errat);
         int skip = eclass == rb_eSysStackError;
+	const int threshold = 1000000000;
+	int width = ((int)log10((double)(len > threshold ?
+					 ((len - 1) / threshold) :
+					 len - 1)) +
+		     (len < threshold ? 0 : 9) + 1);
 
 #define TRACE_MAX (TRACE_HEAD+TRACE_TAIL+5)
 #define TRACE_HEAD 8
 #define TRACE_TAIL 5
 
 	for (i = 1; i < len; i++) {
-	    VALUE line = RARRAY_AREF(errat, i);
+	    VALUE line = RARRAY_AREF(errat, reverse ? len - i : i);
 	    if (RB_TYPE_P(line, T_STRING)) {
-		warn_print_str(rb_sprintf("\tfrom %"PRIsVALUE"\n", line));
+		VALUE bt = rb_str_new_cstr("\t");
+		if (reverse) rb_str_catf(bt, "%*ld: ", width, len - i);
+		write_warn_str(str, rb_str_catf(bt, "from %"PRIsVALUE"\n", line));
 	    }
 	    if (skip && i == TRACE_HEAD && len > TRACE_MAX) {
-		warn_print_str(rb_sprintf("\t ... %ld levels...\n",
+		write_warn_str(str, rb_sprintf("\t ... %ld levels...\n",
 					  len - TRACE_HEAD - TRACE_TAIL));
 		i = len - TRACE_TAIL;
 	    }
 	}
     }
-  error:
-    TH_POP_TAG();
-    th->errinfo = errinfo;
-    rb_thread_raised_set(th, raised_flag);
 }
 
 void
-ruby_error_print(void)
+rb_error_write(VALUE errinfo, VALUE emesg, VALUE errat, VALUE str, VALUE highlight, VALUE reverse)
 {
-    error_print(GET_THREAD());
+    volatile VALUE eclass;
+
+    if (NIL_P(errinfo))
+	return;
+
+    if (errat == Qundef) {
+	errat = Qnil;
+    }
+    eclass = CLASS_OF(errinfo);
+    if (NIL_P(reverse) || NIL_P(highlight)) {
+	VALUE tty = (VALUE)rb_stderr_tty_p();
+	if (NIL_P(reverse)) reverse = tty;
+	if (NIL_P(highlight)) highlight = tty;
+    }
+    if (reverse) {
+	static const char traceback[] = "Traceback "
+	    "(most recent call last):\n";
+	const int bold_part = rb_strlen_lit("Traceback");
+	char buff[sizeof(traceback)+sizeof(bold)+sizeof(reset)-2], *p = buff;
+	const char *msg = traceback;
+	long len = sizeof(traceback) - 1;
+	if (highlight) {
+#define APPEND(s, l) (memcpy(p, s, l), p += (l))
+	    APPEND(bold, sizeof(bold)-1);
+	    APPEND(traceback, bold_part);
+	    APPEND(reset, sizeof(reset)-1);
+	    APPEND(traceback + bold_part, sizeof(traceback)-bold_part-1);
+#undef APPEND
+	    len = p - (msg = buff);
+	}
+	write_warn2(str, msg, len);
+	print_backtrace(eclass, errat, str, TRUE);
+	print_errinfo(eclass, errat, emesg, str, highlight!=0);
+    }
+    else {
+	print_errinfo(eclass, errat, emesg, str, highlight!=0);
+	print_backtrace(eclass, errat, str, FALSE);
+    }
+}
+
+VALUE rb_get_message(VALUE exc);
+
+void
+rb_ec_error_print(rb_execution_context_t * volatile ec, volatile VALUE errinfo)
+{
+    volatile int raised_flag = ec->raised_flag;
+    volatile VALUE errat = Qundef;
+    volatile VALUE emesg = Qundef;
+
+    if (NIL_P(errinfo))
+	return;
+    rb_ec_raised_clear(ec);
+
+    EC_PUSH_TAG(ec);
+    if (EC_EXEC_TAG() == TAG_NONE) {
+	errat = rb_get_backtrace(errinfo);
+    }
+    if (emesg == Qundef) {
+	emesg = Qnil;
+	emesg = rb_get_message(errinfo);
+    }
+
+    rb_error_write(errinfo, emesg, errat, Qnil, Qnil, Qnil);
+
+    EC_POP_TAG();
+    ec->errinfo = errinfo;
+    rb_ec_raised_set(ec, raised_flag);
 }
 
 #define undef_mesg_for(v, k) rb_fstring_cstr("undefined"v" method `%1$s' for "k" `%2$s'")
@@ -250,9 +326,9 @@ static int
 error_handle(int ex)
 {
     int status = EXIT_FAILURE;
-    rb_thread_t *th = GET_THREAD();
+    rb_execution_context_t *ec = GET_EC();
 
-    if (rb_threadptr_set_raised(th))
+    if (rb_ec_set_raised(ec))
 	return EXIT_FAILURE;
     switch (ex & TAG_MASK) {
       case 0:
@@ -260,32 +336,32 @@ error_handle(int ex)
 	break;
 
       case TAG_RETURN:
-	error_pos();
+	error_pos(Qnil);
 	warn_print("unexpected return\n");
 	break;
       case TAG_NEXT:
-	error_pos();
+	error_pos(Qnil);
 	warn_print("unexpected next\n");
 	break;
       case TAG_BREAK:
-	error_pos();
+	error_pos(Qnil);
 	warn_print("unexpected break\n");
 	break;
       case TAG_REDO:
-	error_pos();
+	error_pos(Qnil);
 	warn_print("unexpected redo\n");
 	break;
       case TAG_RETRY:
-	error_pos();
+	error_pos(Qnil);
 	warn_print("retry outside of rescue clause\n");
 	break;
       case TAG_THROW:
 	/* TODO: fix me */
-	error_pos();
+	error_pos(Qnil);
 	warn_print("unexpected throw\n");
 	break;
       case TAG_RAISE: {
-	VALUE errinfo = th->errinfo;
+	VALUE errinfo = ec->errinfo;
 	if (rb_obj_is_kind_of(errinfo, rb_eSystemExit)) {
 	    status = sysexit_status(errinfo);
 	}
@@ -294,17 +370,17 @@ error_handle(int ex)
 	    /* no message when exiting by signal */
 	}
 	else {
-	    error_print(th);
+	    rb_ec_error_print(ec, errinfo);
 	}
 	break;
       }
       case TAG_FATAL:
-	error_print(th);
+	error_print(ec);
 	break;
       default:
 	unknown_longjmp_status(ex);
 	break;
     }
-    rb_threadptr_reset_raised(th);
+    rb_ec_reset_raised(ec);
     return status;
 }
